﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net.WebSockets;
using System.ServiceModel.Syndication;
using System.Threading.Tasks;
using System.Timers;
using System.Xml;
using Discord;
using Discord.WebSocket;

namespace DiscoBot.rss
{
    public class Rss : IModule
    {
        string IModule.Name { get => "Rss"; set => throw new NotImplementedException(); }
        private SocketGuild guild;
        private RssContext rssContext;
        private Dictionary<string, Timer> timers = new Dictionary<string, Timer>();
        private List<WebSocket> webSockets = new List<WebSocket>();

        public Dictionary<string, Func<SocketMessage, string[], Task>> Commands { get; set; } = new Dictionary<string, Func<SocketMessage, string[], Task>>();

        public Rss(SocketGuild guild)
        {
            this.guild = guild;
            rssContext = new RssContext(guild.Id);
            rssContext.Database.EnsureCreated();
            foreach(var f in rssContext.RssFeeds)
            {
                InitializeFeed(f);
            }

            Commands.Add("test",HandleTestCommand);
            Commands.Add("rssadd", HandleRssAddCommand);
            Commands.Add("rssdel", HandleRssDelCommand);
            Commands.Add("rsslist", HandleRssListCommand);
            Commands.Add("rsslistall", HandleRssListAllCommand);
        }

        private Task InitializeFeed(RssFeed feed)
        {
            Console.WriteLine("Found feed " + feed.Name);
            Timer timer = new Timer(feed.CheckInterval.TotalMilliseconds);
            timer.AutoReset = true;
            timer.Elapsed += async (sender, e) =>
            {
                try
                {
                    await Task.Run(() => HandleFeedCheck(feed));
                } catch(Exception)
                {
                    Console.WriteLine("rss: EXCEPTION IN FEED READ!");
                }
            };
            timer.Start();
            timers.Add(feed.Name, timer);
            return Task.CompletedTask;
        }

        private Task DeinitializeFeed(RssFeed feed)
        {
            Timer t = timers[feed.Name];
            t.Stop();
            timers.Remove(feed.Name);
            return Task.CompletedTask;
        }

        private Task HandleFeedCheck(RssFeed feed)
        {
            SocketTextChannel c = guild.Channels.Where(g => g.Id == feed.Channel).Single() as SocketTextChannel;
            try
            {
                XmlReader reader = XmlReader.Create(feed.Url);
                SyndicationFeed f = SyndicationFeed.Load(reader);
                DateTimeOffset newestItem = feed.LastChecked;
                foreach(var item in f.Items)
                {
                    if (item.LastUpdatedTime <= feed.LastChecked)
                        continue;
                    Console.WriteLine("{0}: {1} | {2}", feed.Name, item.LastUpdatedTime, feed.LastChecked);
                    List<string> m = new List<string>();
                    m.Add("<" + feed.Name + "> " + item.Title.Text);
                    if(item.LastUpdatedTime > newestItem)
                    {
                        Console.WriteLine("^-- UPDATE!");
                        newestItem = item.LastUpdatedTime;
                    }
                    foreach (var l in item.Links)
                    {
                        m.Add(l.Uri.ToString());
                    }
                    c.SendMessageAsync(string.Join(" | ", m));
                }
                feed.LastChecked = newestItem;
                rssContext.SaveChangesAsync();
            } catch(System.Net.WebException e)
            {
                c.SendMessageAsync("<" + feed.Name + "> " + "NetworkFailure: " + e.Message);
            } catch(XmlException e)
            {
                c.SendMessageAsync("<" + feed.Name + "> " + "Malformed Response: " + e.Message);
            }
            return Task.CompletedTask;
        }

        private Task HandleRssAddCommand(SocketMessage msg, string[] parameters)
        {
            var gchan = msg.Channel as IGuildChannel;
            string name = parameters[1];
            string url = parameters[2];
            try
            {
                int timeSec = Int32.Parse(parameters[3]);
                RssFeed feed = new RssFeed();
                rssContext.RssFeeds.Add(feed);
                feed.Name = name;
                feed.Url = url;
                feed.Channel = gchan.Id;
                feed.CheckInterval = new TimeSpan(0, 0, timeSec);
                feed.LastChecked = DateTimeOffset.Now;
                try
                {
                    rssContext.SaveChanges();
                    InitializeFeed(feed);
                    msg.Channel.SendMessageAsync("Feed " + feed.Name + " with url " + feed.Url + " saved.");
                }
                catch (InvalidOperationException)
                {
                    msg.Channel.SendMessageAsync("Unable to save rss feed.");
                }
            }
            catch (FormatException)
            {
                msg.Channel.SendMessageAsync("Unable to save rss feed. Invalid check time.");
            }
            
            return Task.CompletedTask;
        }

        private Task HandleRssListCommand(SocketMessage msg, string[] parameters)
        {
            List<RssFeed> feeds = rssContext.RssFeeds.Where(f => f.Channel == msg.Channel.Id).ToList();
            List<string> m = new List<string>();
            m.Add("Feeds for this channel are:");
            m.Add("Name | URL | CheckInterval | LastChecked");
            msg.Channel.SendMessageAsync(string.Join("\n", m));
            m = new List<string>();
            int cnt = 0;
            foreach (var f in feeds)
            {
                m.Add("- " + f.Name + " | " + f.Url + " | " + f.CheckInterval + " | " + f.LastChecked);
                cnt++;
                if(cnt > 4)
                {
                    cnt = 0;
                    msg.Channel.SendMessageAsync(string.Join("\n", m));
                }
            }
            if(cnt > 0)
                msg.Channel.SendMessageAsync(string.Join("\n", m));
            return Task.CompletedTask;
        }

        private Task HandleRssListAllCommand(SocketMessage msg, string[] parameters)
        {
            List<RssFeed> feeds = rssContext.RssFeeds.ToList();
            List<string> m = new List<string>();
            m.Add("Feeds for this guild are:");
            m.Add("Name | URL | Channel | CheckInterval | LastChecked");
            msg.Channel.SendMessageAsync(string.Join("\n", m));
            m = new List<string>();
            int cnt = 0;
            foreach (var f in feeds)
            {
                m.Add("- " + f.Name + " | " + f.Url + " | " + guild.Channels.Where(c => c.Id == f.Channel).Single() + " | " + f.CheckInterval + " | " + f.LastChecked);
                cnt++;
                if (cnt > 4)
                {
                    cnt = 0;
                    msg.Channel.SendMessageAsync(string.Join("\n", m));
                }
            }
            if(cnt > 0)
                msg.Channel.SendMessageAsync(string.Join("\n", m));
            return Task.CompletedTask;
        }

        private Task HandleRssDelCommand(SocketMessage msg, string[] parameters)
        {
            var gchan = msg.Channel as IGuildChannel;
            string name = parameters[1];
            try
            {
                RssFeed f = rssContext.RssFeeds.Where(f => f.Name == name).Single();
                rssContext.Remove(f);
                rssContext.SaveChanges();
                DeinitializeFeed(f);
                msg.Channel.SendMessageAsync("Removed feed " + f.Name);
            } catch(InvalidOperationException)
            {
                msg.Channel.SendMessageAsync("Could not find feed " + name);
            }
            return Task.CompletedTask;
        }

        private Task HandleTestCommand(SocketMessage msg, string[] parameters)
        {
            Console.WriteLine("Handling test command!");
            string para = string.Join(",", parameters);
            msg.Channel.SendMessageAsync("Test succeeded. Params: "+para);
            return Task.CompletedTask;
        }

        public void Initialize()
        {
            Console.WriteLine("Initializing rss...");
        }

        public async Task OnNewWebSocketAsync(WebSocket ws, TaskCompletionSource<object> tcs)
        {
            Console.WriteLine("Calendar " + guild.Id + " has a new websocket.");
            webSockets.Add(ws);
        }
    }
}
